%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
%     G U A R A N T E E D    O U T L I E R    R E M O V A L
%            F O R   R O T A T I O N   S E A R C H
%
%
% This package contains the source code which implements the
% guaranteed outlier removal for rotation search proposed in
% Alvaro PARRA BUSTOS, Tat-Jun CHIN
% Guaranteed Outlier Removal for Rotation Search
% In International Conference on Computer Vision (ICCV) Dec 2015, Santiago
%
% Copyright (c) 2015 Alvaro PARRA BUSTOS (aparra@cs.adelaide.edu.au.)
% School of Computer Science, The University of Adelaide, Australia
% The Australian Center for Visual Technologies
% http://cs.adelaide.edu.au/~aparra
% Please acknowledge the authors by citing the above paper in any academic
% publications that have made use of this package or part of it.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


% This demo contains sample experiments on 3 types of data/applications:
% 1. Synthetic point correspondences.
% 2. Point cloud registration.
% 3. Image stitching.


clear; clc;
close all;
addpath data common;

% Flag to choose between compile the code or use provided binary.
USE_BINARY=1;

if USE_BINARY
    %% Test binary code
    fprintf('*****************************************************\n');
    fprintf('*     Running demo using provided binary            *\n');
    fprintf('*****************************************************\n\n');
    
    try
        % Load point matches
        load( sprintf('ex1_data%d.mat', 100) );
        [~,~,~]=gore(X',Y',th,0,0);
    catch exception
        fprintf('Failed execution of provided binary:\n');
        fprintf('   %s\n\n',exception.message);
        fprintf('Compiling source code...\n');
        compile
    end
    
else
    %% Compile GORE
    fprintf('*****************************************************\n');
    fprintf('*     Running demo compiling code                   *\n');
    fprintf('*****************************************************\n\n');
    
    compile
end

results=struct();

%% Example 1: Synthesized data
% Point matches were synthesized with an inlier ratio of 50%.

fprintf('Example 1: Testing GORE for synthesized data.\n');
fprintf('~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n\n');

dataSize=[100,250,500];

% Startup. First matlab execution takes longer...
load('ex1_data100.mat');
[~,~,~] = gore(X',Y',th,0,0);


for i=1:length(dataSize)
    fprintf('Running GORE for %d point matches.\n',dataSize(i));
    
    % Load point matches
    load( sprintf('ex1_data%d.mat', dataSize(i)) );
    
    % Plot input points
    inlrsIdx = anginlrs(X*R_opt',Y,th);
    figure
    subplot(1,3,1);
    
    X = pointCloud(X);
    Y = pointCloud(Y);
    pcsetcolor(X,'m');
    pcsetcolor(Y,'b');
    
    pcshow(X,'MarkerSize',150)
    hold on
    pcshow(Y,'MarkerSize',150)
    for j=1:X.Count
        x = X.Location(j,:);
        y = Y.Location(j,:);
        if inlrsIdx(j)
            plot3([x(1) y(1)], [x(2) y(2)], [x(3) y(3)],'-g','LineWidth',1)
        else
            plot3([x(1) y(1)], [x(2) y(2)], [x(3) y(3)],'-r','LineWidth',1)
        end
    end
    title('Point matches')
        
    % Execute GORE
    lwbnd=0;
    g = struct();
    tic
    [g.H, g.R, g.lwbnd] = gore(X.Location',Y.Location',th,lwbnd,0);
    g.time = toc;
       
    
    fprintf('Number of point matches...... %d\n', dataSize(i));
    fprintf('Time (s)..................... %f\n', g.time);
    fprintf('Number of removed outliers... %d/%d\n', dataSize(i)-length(g.H), dataSize(i)-opt);
    fprintf('Output lower bound........... %d\n', g.lwbnd);
    fprintf('Optimal solution............. %d\n', opt);
    fprintf('Output rotation matrix:');
    display(g.R);
    
    
    % Plot found rotation over all points.
    subplot(1,3,2);
    %plot_pts_and_matches(X*g.R' ,Y, X*g.R' ,Y, inlrsIdx)
    tform = affine3d([g.R' [0 0 0]';0 0 0 1]);
    X2 = pctransform(X,tform);
    pcshow(X2,'MarkerSize',150)
    hold on
    pcshow(Y,'MarkerSize',150);
    for j=1:X2.Count
        x = X2.Location(j,:);
        y = Y.Location(j,:);
        if inlrsIdx(j)
            plot3([x(1) y(1)], [x(2) y(2)], [x(3) y(3)],'-g','LineWidth',1)
        else
            plot3([x(1) y(1)], [x(2) y(2)], [x(3) y(3)],'-r','LineWidth',1)
        end
    end
    
    title('GORE rotation')
    
    
    % Plot found rotation after removing outliers.
    subplot(1,3,3);    
    pcshow(X2, 'MarkerSize', 150);
    hold on
    pcshow(Y, 'MarkerSize', 150)

    for j=g.H'
        x = X2.Location(j,:);
        y = Y.Location(j,:);
        if inlrsIdx(j)
            plot3([x(1) y(1)], [x(2) y(2)], [x(3) y(3)],'-g','LineWidth',1)
        else
            plot3([x(1) y(1)], [x(2) y(2)], [x(3) y(3)],'-r','LineWidth',1)
        end
    end

    title('GORE rotation without found outliers')
    drawnow
    
end


%% Example 2: Point cloud registration one object from the Stanford dataset.
fprintf('Example 2: Point cloud registration.\n');
fprintf('~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n\n');

% PFH
pfhth=15;

% Angular threshold
th=0.5*pi/180;
numOfMatches=250;

% Load rotation used when computing keypoints. 
rotations = load('ex2_randrot.mat');
rotations = rotations.rotations;

% Load points
load('ex2_data.mat');

% Apply random initialization (also used when obtaining keypoints).
R0=reshape(rotations(1,:),3,3);
X=(R0*X')';

% Load keypoints
load('ex2_keypts');

% Keypoints matching
[X_match,Y_match,~,~,~,~]=matchkeypoints(Mkp,Bkp,Mkp_desc,Bkp_desc,pfhth);

if size(X_match,1)>numOfMatches
    X_match=X_match(1:numOfMatches, :);
    Y_match=Y_match(1:numOfMatches, :);
else
    error('insuficiet matches');
end

% Plot initialisation
load('ex2_opt.mat');
inlrsIdx = anginlrs(X_match*R_opt',Y_match,th);

X = pointCloud(X);
Y = pointCloud(Y);
X_match = pointCloud(X_match);
Y_match = pointCloud(Y_match);

pcsetcolor(X,'m');
pcsetcolor(X_match,'m');

pcsetcolor(Y,'b');
pcsetcolor(Y_match,'b');

figure
subplot(1,3,1)

pcshow(X)
hold on
pcshow(Y)
for i=1:X_match.Count
    x = X_match.Location(i,:);
    y = Y_match.Location(i,:);
    if inlrsIdx(i)
        plot3([x(1) y(1)], [x(2) y(2)], [x(3) y(3)],'-g','LineWidth',2)
    else
        plot3([x(1) y(1)], [x(2) y(2)], [x(3) y(3)],'-r','LineWidth',2)
    end
end
title('input')

g_pcr=struct();

tic
[g_pcr.H, g_pcr.R, g_pcr.lwbnd] = gore(...
    X_match.Location',Y_match.Location',th,0,1);
g_pcr.time = toc;

fprintf('Number of point matches...... %d\n', numOfMatches);
fprintf('Time (s)..................... %f\n', g_pcr.time);
fprintf('Number of removed outliers... %d/%d\n', numOfMatches-length(g_pcr.H), numOfMatches-opt);
fprintf('Output lower bound........... %d\n', g_pcr.lwbnd);
fprintf('Optimal solution............. %d\n', opt);
fprintf('Output rotation matrix:');
display(g_pcr.R);
results.pcr = g_pcr;


% Plot GORE rotation
subplot(1,3,2)
tform = affine3d( [ g_pcr.R' [0 0 0]'; 0 0 0 1 ] );
pcshow(pctransform(X, tform ) )
hold on
pcshow(Y)
title('GORE rotation')

% Plot remaining outliers translated
subplot(1,3,3)
tform = affine3d( [ g_pcr.R' [0 0 0]'; 80 0 0 1 ] );
pcshow(pctransform(X, tform))
hold on
pcshow(Y)
X_match2 = pctransform(X_match, tform);
for i=g_pcr.H'
    x = X_match2.Location(i,:);
    y = Y_match.Location(i,:);
    if inlrsIdx(i)
        plot3([x(1) y(1)], [x(2) y(2)], [x(3) y(3)],'-g','LineWidth',2)
    else
        plot3([x(1) y(1)], [x(2) y(2)], [x(3) y(3)],'-r','LineWidth',2)
    end
end
title('output')
drawnow

%% Example 3: Image stitching
fprintf('Example 3: Image stitching.\n');
fprintf('~~~~~~~~~~~~~~~~~~~~~~~~~~~\n\n');

th=.05*pi/180;

% Read in images
filename1='a.jpg';
filename2='b.jpg';
scale = .25;
I1 = imresize(imread(filename1),scale);
I2 = imresize(imread(filename2),scale);

% Image size
[h,w,~] = size(I1);

% Load SIFT feature points
load('ex3_featurepts.mat');

% Build calibration matrices
K1 = kmatrix(filename1, I1);
K2 = kmatrix(filename2, I2);

% Rotation matrices
Q1 = (inv(K1)*[p1; ones(1,size(p1,2))])';
Q2 = (inv(K2)*[p2; ones(1,size(p2,2))])';
Q1 = normr(Q1);
Q2 = normr(Q2);
 
% Load optimal solution
load('ex3_opt.mat');
inlrsIdx = anginlrs(Q1*R_opt',Q2,th);

% Execute GORE
g_s = struct();
tic
[g_s.H, g_s.R, g_s.lwbnd] = gore(Q1',Q2',th,0,1);
g_s.time = toc;

fprintf('Number of point matches...... %d\n', size(Q1,1));
fprintf('Time (s)..................... %f\n', g_s.time);
fprintf('Number of removed outliers... %d/%d\n', size(Q1,1)-length(g_s.H), size(Q1,1)-opt);
fprintf('Output lower bound........... %d\n', g_s.lwbnd);
fprintf('Optimal solution............. %d\n', opt);
fprintf('Output rotation matrix:');
display(g_s.R);


% Infinite homography
H=K2*g_s.R*inv(K1);

% Plot input
figure, imshow([I1,I2])
hold on
for i=1:size(p1,2)
    p1i = p1(:,i);
    p2i = p2(:,i);
    if inlrsIdx(i)
        plot([p1i(1), p2i(1)+w],[p1i(2), p2i(2) ],'-g','linewidth',1)
    else
        plot([p1i(1), p2i(1)+w],[p1i(2), p2i(2) ],'-r','linewidth',1)
    end
end


% Plot matches after removing detected outliers
figure, imshow([I1,I2])
hold on

for i=g_s.H'
    p1i = p1(:,i);
    p2i = p2(:,i);
    if inlrsIdx(i)
        plot([p1i(1), p2i(1)+w],[p1i(2), p2i(2) ],'-g','linewidth',1)
    else
        plot([p1i(1), p2i(1)+w],[p1i(2), p2i(2) ],'-r','linewidth',1)
    end
end

% Plot blended images
I1p = paddedimage(I1,.25);

imgwarped  = imagewarping(I1p,I2,H);
imgblended = imageblending(I1p,imgwarped);

figure;
imshow(imgblended);

