///////////////////////////////////////////////////////////////////////////
//
//     G U A R A N T E E D    O U T L I E R    R E M O V A L
//            F O R   R O T A T I O N   S E A R C H
//
//
// This package contains the source code which implements the
// guaranteed outlier removal for rotation search proposed in
// Alvaro PARRA BUSTOS, Tat-Jun CHIN
// Guaranteed Outlier Removal for Rotation Search
// In International Conference on Computer Vision (ICCV) Dec 2015, Santiago
//
// Copyright (c) 2015 Alvaro PARRA BUSTOS (aparra@cs.adelaide.edu.au.)
// School of Computer Science, The University of Adelaide, Australia
// The Australian Center for Visual Technologies
// http://cs.adelaide.edu.au/~aparra
// Please acknowledge the authors by citing the above paper in any academic
// publications that have made use of this package or part of it.
///////////////////////////////////////////////////////////////////////////


#include <vector>
#include "matrix.h"
#include "mex.h"
#include "math.h"

//Windows
#include <algorithm> //for sort
#include <cmath>
#ifndef M_PI
#define M_PI 3.14159265358979323846
#endif

#define DUMMY_PRECISION 1e-12
#define PI  M_PI
#define TWOPI  2*M_PI
#define MAX(a,b) ((a) > (b) ? a : b)
#define MIN(a,b) ((a) < (b) ? a : b)


/**
 * @brief transpose Compute the transpose of a rotation matrix.
 * @param R Rotation matrix.
 * @return A pointer to the rotation matrix after it is transposed.
 */
inline
double *transpose(double *R)
{
    double tmp;
    tmp = R[1]; R[1]=R[3]; R[3]=tmp; //swap
    tmp = R[2]; R[2]=R[6]; R[6]=tmp;
    tmp = R[5]; R[5]=R[7]; R[7]=tmp;
    return R;
}


/**
 * @brief Create a new 3xn matrix such that columns of X are normalised..
 * @param X 3xn matrix where each column is a 3D point.
 * @param cols Number of points.
 * @return Pointer to the new normalised matrix.
 */
inline
double *normalisecols(const double *X, const size_t n)
{
    double *outArray = (double *)mxMalloc(3*n*sizeof(double));
    
    double norm;
    int i,j;
    for(j=0; j<n; j++) /* Compute a matrix with normalized columns */
    {
        norm = 0.0;
        for(i=0; i<3; i++)
        {
            norm += (X[i + 3*j])*(X[i + 3*j]);
        }
        norm = sqrt(norm);
        for(i=0; i<3; i++)
        {
            outArray[i + 3*j] = X[i + 3*j]/norm;
        }
    }
    return outArray;
}



/**
 * @brief rotationFromAngle Obtain rotation matrix about the z-axis.
 * @param theta Rotation angle.
 * @param R Rotation matrix about the z-axis with rotation angle theta.
 */
inline
void rotationFromAngle(const double theta, double *R)
{
    const double s = sin(theta);
    R[0]=R[4]=cos(theta);
    R[1]=s;
    R[2]=R[5]=R[6]=R[7]=0.0;
    R[3]=-s;
    R[8]=1.0;
}


/**
 * @brief Obtain a rotation matrix that aligns a with the z-axis.
 * @param a 3D unit-norm vector.
 * @param R Obtained rotation matrix.
 * @return False if c is too close to the [0 0 -1] where the algorithm is
 * undefined.
 */
inline
bool rotationFromVector(const double *a, double *R)
{
    const double &c = a[2]; //dot product with north pole [0 0 1]
    
    if (c < -1.0 + DUMMY_PRECISION)
    {
        return false;
    }
    
    const double s = sqrt( (1.0+c)*2.0 );
    const double invs = 1.0/s;
    const double x = a[1]*invs;
    const double y = -a[0]*invs;
    const double w = s*0.5;
    
    //Convert quaternion to rotation matrix
    const double tx  = 2.0*x;
    const double ty  = 2.0*y;
    const double twx = tx*w;
    const double twy = ty*w;
    const double txx = tx*x;
    const double tyy = ty*y;
    
    R[0] = 1.0-tyy;
    R[1] = R[3] = ty*x;
    R[2] = -twy;
    R[4] = 1.0-txx;
    R[5] = twx;
    R[6] = twy;
    R[7] = -twx;
    R[8] = 1.0-(txx+tyy);
    
    return true;
}


// Compute T = R*S
// R,T,S are 3x3 matrices
inline
void matrix3Multiply(double *T, double *R, double *S)
{
    T[0]=R[0]*S[0]+R[3]*S[1]+R[6]*S[2]; T[3]=R[0]*S[3]+R[3]*S[4]+R[6]*S[5]; T[6]=R[0]*S[6]+R[3]*S[7]+R[6]*S[8];
    T[1]=R[1]*S[0]+R[4]*S[1]+R[7]*S[2]; T[4]=R[1]*S[3]+R[4]*S[4]+R[7]*S[5]; T[7]=R[1]*S[6]+R[4]*S[7]+R[7]*S[8];
    T[2]=R[2]*S[0]+R[5]*S[1]+R[8]*S[2]; T[5]=R[2]*S[3]+R[5]*S[4]+R[8]*S[5]; T[8]=R[2]*S[6]+R[5]*S[7]+R[8]*S[8];
}

inline
void rotatePoint(double *y, double *R, double *x)
{
    y[0] = R[0]*x[0] + R[3]*x[1] + R[6]*x[2];
    y[1] = R[1]*x[0] + R[4]*x[1] + R[7]*x[2];
    y[2] = R[2]*x[0] + R[5]*x[1] + R[8]*x[2];
}


inline
int eval(double *R, double *X, double *Y, double *Rx, double *Ry,
         double *D, long int n,
         const double th, const int lwbnd)
{
    mxAssert(n>0,"");
    mxAssert(th>0,"");
    
    double d;
    double x[3], y;
    
    int q=0; //number of matches
    // Squared Euclidean threshold
    const double eTh = 4.0*sin(0.5*th)*sin(0.5*th);
    
    for(int i=0; i<n && q+(n-i)>lwbnd; i++)
    {
        //rotate x
        rotatePoint(x, Rx, X+3*i);
        
        const double* ypr = Y+3*i;
        
        y = Ry[0]*ypr[0] + Ry[3]*ypr[1] + Ry[6]*ypr[2];
        d = R[0]*x[0] + R[3]*x[1] + R[6]*x[2] - y;
        D[i] = d*d;
        if (D[i]>eTh)
        {
            continue;
        }
        
        y = Ry[1]*ypr[0] + Ry[4]*ypr[1] + Ry[7]*ypr[2];
        d = R[1]*x[0] + R[4]*x[1] + R[7]*x[2] - y;
        D[i] += d*d;
        if (D[i]>eTh)
        {
            continue;
        }
        
        y = Ry[2]*ypr[0] + Ry[5]*ypr[1] + Ry[8]*ypr[2];
        d = R[2]*x[0] + R[5]*x[1] + R[8]*x[2] - y;
        D[i] += d*d;
        
        if (D[i]<=eTh)
        {
            q++;
        }
    }
    
    return q;
}



/**
 * @brief feasibleRegrionAngles return upper and lower angles to bound the
 * uncertainty region.
 * @param m Inclination in the xy-plane: tan(th).
 * @param z Z coordinate of x.
 * @param xinc Inclination of x.
 * @param cosDazi Cosine of the azimuthal difference between x and y.
 * @param th Angular threshold.
 * @param u Output angle bounding the upper region.
 * @param l Output angle bounding the lower region.
 */
void feasibleRegrionAngles(double m, double z, double xinc, double cosDazi,
                           double th, double &u, double &l)
{
    mxAssert(m>=-1 && m<=1,"");     // tan(th) --> inclination of the rect
    mxAssert(z>=-1 && z<=1,"");     // z coordinate of m
    mxAssert(xinc>=0 && z<=PI,"");  // inclination angle of m
    mxAssert(cosDazi>=-1 && cosDazi<=1,"");//cos of azim diff between m and b
    
    //-----------------------------------
    // Border conditions
    //-----------------------------------
    
    // Same azimuth
    if (fabs(cosDazi-1.0)<DUMMY_PRECISION)
    {
        l=u=0;
        return;
    }
    
    // Oposite azimuth
    if (fabs(cosDazi+1.0)<DUMMY_PRECISION)
    {
        l=u=2*th;
        if(xinc-u<0)
        {
            u=xinc;
        }
        if(xinc+l>PI)
        {
            l=PI-xinc;
        }
        return;
    }
    
    double d = sqrt(1.0-z*z)*cosDazi; // rad. of circ defiend by x
    double md = m*d;
    double c = 1.0 + m*m; //denom quad equation
    double x;
    
    //-----------------------------------
    // Upper region
    //-----------------------------------
    
    // Solve intersection x
    x = ( m*(md-z) + sqrt(c-(z-md)*(z-md)) ) / c;
    
    // Compute theta (polar coordinate) and u = xinc-theta
    if(x>0)
    {
        if (x > d-z/m) // y>0
        {
            // theta = asin(x)
            mxAssert(asin(x)>=0 && asin(x) <= xinc+DUMMY_PRECISION,"");
            u = MAX(xinc-asin(x), 0.0);
        }
        else
        {
            // theta = PI-asin(x)
            mxAssert((PI-asin(x))>=0 && (PI-asin(x)) <= xinc+DUMMY_PRECISION,"");
            u = MAX(xinc-(PI-asin(x)), 0.0);
        }
    }
    else
    {
        // theta=0;
        u=xinc;
    }
    
    //-----------------------------------
    // Lower region
    //-----------------------------------
    
    // Solve intersection for (-m)
    x = ( m*(md+z) + sqrt(c -(z+md)*(z+md)) ) / c;
    
    // Compute theta and l = theta-xinc
    if(x>0)
    {
        if (x<d+z/m) // y>0
        {
            // theta = asin(x)
            mxAssert(asin(x)>=0 && asin(x)>=xinc-DUMMY_PRECISION,"");
            l = MAX(asin(x)-xinc, 0.0);
        }
        else
        {
            // theta = PI-asin(x)
            mxAssert( (PI-asin(x)) >=0 && (PI-asin(x))>=xinc-DUMMY_PRECISION,"");
            l = MAX(PI-asin(x)-xinc, 0.0);
        }
    }
    else
    {
        // theta = PI
        l = PI-xinc;
    }
    
    mxAssert(u>=0 && u<=2*th,"");
    mxAssert(l>=0 && l<=2*th,"");
}


/**
 * @brief bndAlpha Computes alpha angle for (theta-gamma) in the [0,pi]
 * interval.
 * @param rho theta-gamma
 * @param xinc Inclination of x.
 * @param th Angular threshold.
 * @return angle alpha
 */
double bndAlpha(double rho, double xinc, double th)
{
    mxAssert (rho>=0 && rho<=PI,"");
    mxAssert(xinc>=0 && xinc<=PI,"");
    mxAssert(th>=0 && th<=PI/2.0,"");
    
    const double w = 1.0/sin(xinc);
    double m,b;
    if (rho > asin(w*sin(th)) )
    {
        m = -2.0*sin(0.5*th)*w ;
        b = ( 2.0*rho*sin(0.5*th)+th)*w ;
    }
    else
    {
        m = 2.0*sin(0.5*th)*w ;
        b = ( -2.0*rho*sin(0.5*th)+th)*w ;
    }
    
    double alpha = b/(2.0*sqrt(2.0)/PI - m);
    
    if(alpha>=0 && alpha<=PI/4.0)
    {
        mxAssert(alpha>=0 && alpha<=PI/4.0,"");
        return alpha;
    }
    
    // Intersect second segment
    alpha = (sqrt(2.0)-1.0-b)/(m+2.0*(sqrt(2.0)-2.0)/PI);
    
    mxAssert(alpha>=PI/4.0 && alpha<=PI/2.0,"");
    return alpha;
}


/**
 * Bound angle beta for rho in the [0,pi] interval.
 * @brief bndBeta Computes beta angle for (theta+gamma) in the [0,pi]
 * interval.
 * @param rho theta+gamma
 * @param xinc Inclination of x.
 * @param th Angular threshold.
 * @return Beta angle
 */
inline
double bndBeta(double rho, double xinc, double th)
{
    mxAssert(rho>=0 && rho<=PI,"");
    mxAssert(xinc>=0 && xinc<=PI,"");
    mxAssert(th>=0 && th<=PI/2.0,"");
    
    const double w = 1.0/sin(xinc);
    
    const double m = 2.0*sin(0.5*th)*w;
    const double b = (2.0*rho*sin(0.5*th)+th)*w;
    
    double beta = b/(2.0*sqrt(2.0)/PI - m);
    
    mxAssert(beta>=0,"");
    
    if(beta>=0 && beta<=PI/4.0)
    {
        mxAssert(beta>=0 && beta<=PI/4.0,"");
        return beta;
    }
    
    // Intersect second segment
    beta = (sqrt(2.0)-1.0-b)/(m+2.0*(sqrt(2.0)-2.0)/PI);
    
    mxAssert(beta>=PI/4.0 && beta<=PI/2.0,"");
    return beta;
}


/**
 * @brief uncertaintyAngles Find the parametres alpha and beta for the
 * uncertainty interval [theta-alpha, theta+beta].
 * @param gamma Uncertainty angle containing S_eps(y).
 * @param theta Rotation angle about the z axis to align x to y.
 * @param xinc Inclination of x.
 * @param th Angular threshold.
 * @param alpha Left angle defining the uncertainty
 * interval [theta-alpha, theta+beta].
 * @param beta Right angle defining the uncertainty
 * interval [theta-alpha, theta+beta].
 * @return false if angles alpha and beta could not be obtained.
 */
bool uncertaintyAngles(double gamma, double theta, double xinc, double th,
                       double &alpha, double &beta)
{
    mxAssert(gamma>=0 && gamma<=PI,"");
    mxAssert(theta>=0 && theta<=TWOPI,"");
    mxAssert(xinc>=0 && xinc<=PI,"");
    mxAssert(th>=0 && th<=PI/2.0,"");
    
    //------------------------------------------------------
    //   Solve for beta
    //------------------------------------------------------
    double rho = theta+gamma;
    
    if(rho<=PI)
    {
        mxAssert(rho>0,"");
        
        // Check if there is no solution
        if( sin(xinc) <= (th+(2*rho+PI)*sin(0.5*th)) )
        {
            return false;
        }
        
        beta = bndBeta(rho, xinc, th);
        
        // Fix beta if it is grather than the max. possible solution.
        double aux = sin(TWOPI*sin(0.5*th)+th) / sin(xinc);
        mxAssert(aux>=0,"");
        if(aux <= 1.0)
        {
            beta = MIN(beta, asin(aux));
        }
    }
    else // rho>pi
    {
        // Solve as alpha of 2*pi-rho
        rho = TWOPI-rho;
        beta = bndAlpha(rho, xinc, th);
    }
    
    mxAssert(0<=beta && beta <= PI/2.0,"");
    
    //------------------------------------------------------
    //   Solve for alpha
    //------------------------------------------------------
    rho = theta-gamma;
    
    if(rho>0)
    {
        // Check if there is not solution
        if( sin(xinc) <= (th+(2.0*rho+PI)*sin(0.5*th)) )
        {
            return false;
        }
        
        alpha = bndAlpha(rho, xinc, th);
    }
    else
    {
        // Solve as beta for |rho|
        rho = -rho;
        alpha = bndBeta(rho, xinc, th);
    }
    
    mxAssert(0<=alpha && alpha>=asin(sin(th)/sin(xinc)) && alpha <= PI/2.0,"");
    
    return true;
}


/**
 * @brief angularBoundingInterval Find uncertainty interval [l r]
 * for estimate rotation theta.
 * @param theta rotation angle about the z axis to align x to y.
 * @param xinc Inclination of x.
 * @param yinc Inclination of y.
 * @param th Angular threshold.
 * @param l Left value in [l, r].
 * @param r Right value in [l, r].
 * @return false if the interval [l r] could not be obtained.
 */
bool angularBoundingInterval(double theta,double xinc, double yinc,
                             double th, double &l, double &r)
{
    mxAssert(0<=theta && theta<=TWOPI,"");
    mxAssert(xinc>=0 && xinc<=PI,"");
    mxAssert(yinc>=0 && yinc<=PI,"");
    mxAssert(th>=0 && th<=PI/2.0,"");
    
    double alpha, beta;
    const double gamma = asin( sin(th)/sin(yinc) );
    
    mxAssert(gamma >=0 && gamma<=.5*PI, "");
    
    if(theta<PI)
    {
        if(!uncertaintyAngles(gamma, theta, xinc, th, alpha, beta))
        {
            return false;
        }
    }
    else
    {
        if (!uncertaintyAngles(gamma, TWOPI-theta, xinc, th, beta, alpha))
        {
            return false;
        }
    }
    
    l = theta - gamma - alpha;
    r = theta + gamma + beta;
    
    return true;
}


/**
 * @brief Structure to be sorted by the stabbing method.
 */
struct LimitLabel
{
    double value;
    short label;  //label in {1,-1}
};


/**
 * @brief Operator to sort LimitLabel instances.
 */
struct limitLabelComparator
{
    bool operator()(const LimitLabel &v1, const LimitLabel &v2) const
    {
        return  v1.value < v2.value;
    }
};


// Guaranteed Outlier REmoval main algorithm.
// Return the new size of H
int gore(size_t *H, double *X, double *Y,
         int numOfPts, double th, int lwbnd,
         double *Rout, int &outLwbnd)
{
    mxAssert(numOfPts>=lwbnd, "");
    
    const double tanTh = tan(th);
    
    // Euclidean threshold
    const double eTh = 4.0*sin(0.5*th)*sin(0.5*th);
    
    // Uncertainty angle interval limits
    double thetaL, thetaR;
    
    // Aux. vars. to fix angles to lie in [0,2pi]
    double segAEnd, segBSrt;
    
    // Estimated angle in [0,2pi] about the Z axis.
    double theta;
    
    // Dot product of x and y.
    double dot;
    
    // Angular distance in [0,pi].
    double angdist;
    
    // Density in stabbing
    int dsty, maxDsty;
    
    // Counters
    int numberOfLimits, numOfInsertions;
    
    // Rotation matrices
    double Ropt[9], Raux[9], Rx[9], Ry[9];
    
    // Inclinations
    double xinc, yinc;
    
    // Stabbing vars. to control iterations.
    int i, offset;
    
    // Stabbing opt. val.
    double opt;
    
    int optIdx;
    
    // Lower bound for subproblem k.
    int lwbndK;
    
    // Feasible region limits.
    double uth, lth;
    
    // Array to be sorted for the stabbing algorithm
    struct LimitLabel *ll = new struct LimitLabel[4*numOfPts];
    
    // Vector indicating point matches to be tested.
    std::vector<bool> potentialOutlrs(numOfPts, true);
    
    //Euclidean distances
    double *Distance = (double *)mxMalloc(numOfPts*sizeof(double));
    
    double xk[3], yk[3];
    
    // Iterate to try to reject matches.
    for (int k=0; k<numOfPts; )
    {
        // Move to a potential outlier to reject
        while (k<numOfPts && !potentialOutlrs[k])
        {
            k++;
        }
        
        if (k==numOfPts)
        {
            break;
        }
        
        // Find Rx and Ry to align the visited point match with the N. Pole
        if (!rotationFromVector(X+k*3, Rx) || !rotationFromVector(Y+k*3, Ry))
        {
            k++;
            continue;
        }
        
        //-----------------------------------------------------------------
        //  Stabbing on rotation angle about the Z axis
        //-----------------------------------------------------------------
        offset = 1; // Initialise for i==k
        numberOfLimits = numOfInsertions = 0;
        
        
        for (i=0; i<numOfPts &&
             (offset+numOfInsertions+(numOfPts-i))>=lwbnd; i++)
        {
            
            // skip point match k
            if (i==k)
            {
                continue;
            }
            
            //xk = Rx * X[3*i]
            rotatePoint(xk, Rx, X+3*i);
            rotatePoint(yk, Ry, Y+3*i);
            
            
            double &x3 = xk[2];
            double &y3 = yk[2];
            
            
            if (y3>1)
            {
                y3=1.0;
            }
            yinc = acos(y3);
            
            mxAssert(yinc>=0 && yinc<=PI ,"incorrect inclination" );
            
            // Make sure yinc is well defined (Non NAN values)
            if ( (1.0-x3 < DUMMY_PRECISION)  ) //x = North Pole
            {
                // if y touches the pole it also matches x
                if ( 1.0-y3 <= DUMMY_PRECISION || yinc <= th)
                {
                    offset++;
                }
                continue;
            }
            
            if ( x3+1.0 < DUMMY_PRECISION ) //x = South Pole
            {
                // if y touches the south it also matches x
                if ( y3+1.0 <= DUMMY_PRECISION || yinc >= PI-th)
                {
                    offset++;
                }
                continue;
            }
            
            xinc = acos(x3);
            
            // Make sure xinc is well defined (Non NAN values)
            
            // y = North Pole, South Pole
            if ( (1.0-y3 < DUMMY_PRECISION) || (y3+1.0 < DUMMY_PRECISION) )
            {
                offset++;
                continue;
            }
            
            //---------------------------------------------------------------
            // x and y are not in the poles
            //---------------------------------------------------------------
            mxAssert(xinc>0 && xinc<PI,"");
            mxAssert(yinc>0 && yinc<PI,"");
            
            //projections
            double &x1 = xk[0];
            double &x2 = xk[1];
            double &y1 = yk[0];
            double &y2 = yk[1];
            
            
            dot = (x1*y1+x2*y2)/sqrt((x1*x1+x2*x2)*(y1*y1+y2*y2));
            mxAssert(dot>=-1 && dot<=1,"");
            
            feasibleRegrionAngles(tanTh, x3, xinc, dot, th, uth, lth);
            uth+=th; //add patch uncertainty
            lth+=th;
            
            // No solutin if x is out of the feasible region
            if(xinc>yinc)
            {
                if( (xinc-yinc) > lth+th)
                {
                    continue;
                }
            }
            else if( (yinc-xinc) > uth+th)
            {
                continue;
            }
            
            // Check Poles
            if ( ( (xinc-uth-th) <= 0 ) || ( (xinc+lth+th) >= PI ) ||
                ( (yinc-th    ) <= 0 ) || ( (yinc+th    ) >= PI ) )
            {
                offset++;
                continue;
            }
            
            angdist = acos(dot);
            mxAssert(angdist>=0 && angdist<= PI,"");
            
            // Create segments in [0,2pi]
            theta = x1*y2-x2*y1 >=0 ?  angdist: TWOPI - angdist;
            
            if (!angularBoundingInterval(theta,xinc,yinc,th,thetaL,thetaR))
            {
                offset++;
                continue;
            }
            
            if (thetaL<=0 && thetaR>=TWOPI)
            {
                offset++;
                continue;
            }
            
            if (thetaL<0 || thetaR>TWOPI)
            {
                //compute two segments and check if them intersects
                if (thetaL<0 ) // l---[---r     ]
                {
                    segAEnd = thetaR;
                    segBSrt = TWOPI+thetaL;
                }
                else // [    l--]---r
                {
                    segAEnd = thetaR-TWOPI;
                    segBSrt = thetaL;
                }
                
                if (segBSrt<=segAEnd) // seg_a and seg_b intersect
                {
                    offset++;
                }
                else
                {
                    ll[numberOfLimits].value = 0;       ll[numberOfLimits++].label = 1;
                    ll[numberOfLimits].value = segAEnd; ll[numberOfLimits++].label =-1;
                    ll[numberOfLimits].value = segBSrt; ll[numberOfLimits++].label = 1;
                    ll[numberOfLimits].value = TWOPI;   ll[numberOfLimits++].label =-1;
                    numOfInsertions++;
                }
            }
            else //segment_l>0 && segment_r<TWOPI
            {
                ll[numberOfLimits].value = thetaL;  ll[numberOfLimits++].label = 1;
                ll[numberOfLimits].value = thetaR;  ll[numberOfLimits++].label =-1;
                numOfInsertions++;
            }
            
        } // End iterate over intervals
        
        if ( numOfInsertions+offset < lwbnd )
        {
            numOfPts--;
            
            //reorganise memory if k is not the last point
            if (k<numOfPts)
            {
                const int lstPtPtr = 3*numOfPts;
                std::copy(X+lstPtPtr,X+lstPtPtr+3,X+3*k);
                std::copy(Y+lstPtPtr,Y+lstPtPtr+3,Y+3*k);
                H[k]=H[numOfPts];
                potentialOutlrs[k]=potentialOutlrs[numOfPts];
            }
            
            continue;
        }
        
        //-----------------------------------
        // Obtain max. intersection density
        //-----------------------------------
        std::sort(ll, ll+numberOfLimits, limitLabelComparator());
        
        dsty = maxDsty = offset;
        optIdx=0;
        for(i=0; i<numberOfLimits; i++)
        {
            dsty += ll[i].label;
            if(dsty > maxDsty)
            {
                maxDsty = dsty;
                optIdx = i;
            }
        }
        
        //----------------------------------------
        // Estimate and update lower bound
        //----------------------------------------
        opt = .5*(ll[optIdx].value + ll[optIdx+1].value);
        
        rotationFromAngle(opt, Ropt);
        
        lwbndK = eval(Ropt, X, Y, Rx, Ry, Distance, numOfPts, th, lwbnd);
        
        
        // Update lower bound
        if(lwbndK > lwbnd)
        {
            lwbnd = lwbndK;
            
            // Update array of potential outlrs from position k+1
            for (i=k+1; i<numOfPts; i++)
            {
                potentialOutlrs[i] = Distance[i]>eTh;
            }
            
            // Update output rotation matrix: Rout = Ry.transpose()*Ropt*Rx;
            matrix3Multiply(Raux, Ropt, Rx);
            matrix3Multiply(Rout, transpose(Ry), Raux);
        }
        
        //------------------------------------------------
        // Check upper bound estimation
        //------------------------------------------------
        if (maxDsty < lwbnd)
        {
            numOfPts--;
            
            //reorganise memory if k is not the last point
            if (k<numOfPts)
            {
                const int lstPtPtr = 3*numOfPts;
                std::copy(X+lstPtPtr,X+lstPtPtr+3,X+3*k);
                std::copy(Y+lstPtPtr,Y+lstPtPtr+3,Y+3*k);
                H[k]=H[numOfPts];
                potentialOutlrs[k]=potentialOutlrs[numOfPts];
            }
            
            continue;
        }
        
        // Move to the next correspondence to be tested
        k++;
    } // End iteration over correspondences
    
    delete []ll;
    mxFree(Distance);
    
    outLwbnd = lwbnd;
    
    return numOfPts;
}


int goreRep(size_t *H, double *X, double *Y, int numOfPts, double th,
            int lwbnd, double *Rout, int &outLwbnd)
{
    mxAssert(numOfPts>0,"invalid number of points");
    mxAssert(lwbnd>=0,"invalid lower bound");
    mxAssert (th>0,"invalid threshold");
    
    int n = gore(H,X,Y,numOfPts,th,lwbnd,Rout,outLwbnd);
    mxAssert(outLwbnd>=lwbnd, "");
    
    int bestLwbnd = outLwbnd;
    
    double R[9];
    
    while ( n < numOfPts)
    {
        numOfPts = n;
        
        n = gore(H,X,Y,numOfPts,th,bestLwbnd,R,outLwbnd);
        
        //Update the best solution
        if (outLwbnd > bestLwbnd)
        {
            bestLwbnd = outLwbnd;
            std::copy(R,R+9,Rout);
        }
    }
    mxAssert(n==numOfPts, "");
    
    outLwbnd = bestLwbnd;
    
    return numOfPts;
}


int _gore(size_t *H, const double *X, const double *Y, int numOfPts,
             double th, int lwbnd, bool repFlag,
             double *R, int &outLwbnd)
{
    
    mxAssert(numOfPts>=0 , "invalid number of points, it must be >= 0");
    mxAssert(lwbnd>=0 , "invalid input lower bound, it must be > 0");
    mxAssert(th>0 , "invalid threshold, it must be > 0");
    
    
    // Normalise points

    double *Xn = normalisecols(X, numOfPts);
    double *Yn = normalisecols(Y, numOfPts);
    
    int n;
    if (!repFlag)
    {
        //n = _gore_l2(H,X,Y,XLen,YLen,numOfPts,th,TH,lwbnd,R,outLwbnd);
        n = gore(H, Xn,Yn,numOfPts,th,lwbnd,R,outLwbnd);
    }
    else
    {
        //n = _gore_l2_rep(H,X,Y,XLen,YLen,numOfPts,th,TH,lwbnd,R,outLwbnd);
        n = goreRep(H, Xn,Yn,numOfPts,th,lwbnd,R,outLwbnd);
    }

    mxAssert(numOfPts-n>=0, "wrong result");
    mxAssert(outLwbnd>=lwbnd, "wrong result");
    
    mxFree(Xn);
    mxFree(Yn);
    
    return n;
}



// Input
#define X_IN      prhs[0]
#define Y_IN      prhs[1]
#define TH_IN     prhs[2]
#define LWBND_IN  prhs[3]
#define REPEAT_IN prhs[4] //If 1, repeat until no outliers can be removed.
// Output
#define H_OUT     plhs[0]
#define R_OUT     plhs[1]
#define LWBND_OUT plhs[2]
void mexFunction(int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[])
{
    if (nlhs != 3)
    {
        mexErrMsgTxt("Invalid number of output variables. Expected 3 LHS vars.");
    }
    
    if (nrhs != 5)
    {
        mexErrMsgTxt("Invalid number of parameters. Expected 5 parameters.");
    }
    
    const double th = *mxGetPr(TH_IN);
    const int lwbnd = (int)*mxGetPr(LWBND_IN);
    const int numOfPts = (int)mxGetN(X_IN);
    const bool repeat = (bool)*mxGetPr(REPEAT_IN);;
    
    if (!mxIsDouble(X_IN) || !mxIsDouble(Y_IN))
    {
        mexErrMsgTxt("Input argument must be of type double.");
    }
    
    
    if (mxGetM(X_IN) != 3)
    {
        mexErrMsgTxt("Invalid input. Expected 3xn matrices");
    }
    
    if (mxGetM(Y_IN) != 3)
    {
        mexErrMsgTxt("Invalid input. Expected 3xn matrices");
    }
    
    // Allocate output rotation matrix
    R_OUT = mxCreateDoubleMatrix(3, 3, mxREAL);
    
    int outLwbnd;
    
    size_t *H = (size_t *)mxMalloc(numOfPts*sizeof(size_t));
    
    for(int i=0; i<numOfPts; i++)
    {
        H[i]=i;
    }

    int n = _gore(H,mxGetPr(X_IN), mxGetPr(Y_IN),numOfPts,
                                th, lwbnd, repeat, mxGetPr(R_OUT), outLwbnd);
    
    
    mxAssert(outLwbnd>=lwbnd, "wrong result");
    
    
    H_OUT = mxCreateNumericMatrix(n, 1, mxUINT32_CLASS, mxREAL);
    
    uint32_T *hpr = (uint32_T *)mxGetData(H_OUT);
    
    for(int i=0; i<n; i++)
    {
        hpr[i] = (uint32_T) (1+H[i]); //convert to matlab indices
    }
    
    mxFree(H);
    
    LWBND_OUT = mxCreateDoubleScalar(outLwbnd);
}
